#include <stdio.h>
#include "Fifo.h"

Fifo :: Fifo()
{
  astate = 0;	// not allocated
  buffer = NULL;
  totsize = 0;
  datasize = 0;
  start = 0;
}

Fifo :: Fifo(unsigned int size)
{
  buffer = new char[size];
  if (buffer != NULL)
    {
      astate = 1;
      totsize = size;
      datasize = 0;
      start = 0;
    }
  else
    {
      astate = 0;
      totsize = 0;
      datasize = 0;
      start = 0;
    }
}

Fifo :: ~Fifo()
{
  if (astate)
    delete buffer;
}

void Fifo::Flush(){
  astate = 1;
  //totsize = size;
  datasize = 0;
  start = 0;
  //for (int x =0; x < sizeof(buffer); x++) {
  //  buffer[x] = 0;
  //}
}

int Fifo :: ReAlloc(unsigned int size)
{
  if (astate)
    delete buffer;
  buffer = new char[size];
  if (buffer != NULL)
    {
      astate = 1;
      totsize = size;
      datasize = 0;
      start = 0;
    }
  else
    {
      astate = 0;
      totsize = 0;
      datasize = 0;
      start = 0;
      return -1;
    }
  return 0;
}

void * Fifo::Read(void *buf, unsigned int &len)
{
  if (len > datasize)
    len = datasize;
  unsigned int rest;
  if (len > (totsize-start))
    rest = len-(totsize-start);
  else
    rest = 0;
  unsigned int first = len-rest;
  memcpy(buf, buffer+start, first);
  memcpy((char *)buf+first, buffer, rest);
  datasize-=len;
  start+=len;
  if (start >= totsize)
    start = rest;
  //if (datasize == 0) printf("in fifo READ, data is zero\n");
  return buf;
}

int Fifo :: Write(void *buf, unsigned int len)
{
  unsigned int end;
  end = start+datasize;
  if (end > totsize)
    end = end-totsize;
  if (len > (totsize-datasize))
    return -1;
  unsigned int rest;
  if ((len+end) > totsize)
    rest = (len+end)-totsize;
  else
    rest = 0;
  unsigned int first = len-rest;
  memcpy(buffer+end, buf, first);
  memcpy(buffer, (char *)buf+first, rest);
  
  datasize+=len;
  //if (datasize == 0) printf("in fifo WRITE, data is zero\n");
  return len;
}

unsigned int Fifo :: FreeSpace(void)
{
  return (totsize-datasize);
}

unsigned int Fifo :: UsedSpace(void)
{
  return datasize;
}
