
#include <m_pd.h>
#include "ReadMad.h"

#define mad_f_tofloat(x)  ((float) ((x) / (float) (1L << MAD_F_FRACBITS)))

ReadMad::ReadMad() {
  mad_fd = NULL;
}

ReadMad::~ReadMad() {
  if (mad_fd)
    fclose(mad_fd);
  if ( mad_init) {
    mad_synth_finish (&synth);
    mad_frame_finish (&frame);
    mad_stream_finish(&stream);
    mad_init = 0;
  }
  
}				


void ReadMad::fill_buffer( long newoffset) {
  size_t bytes_read;
  if ( seekable && newoffset >= 0) {
    fseek( mad_fd, offset + newoffset, SEEK_SET);
    bytes_read = fread( mad_map, sizeof(unsigned char),MAD_BUFSIZE, mad_fd);
    bytes_avail = bytes_read;
    map_offset = newoffset;
  } else {
    memmove(mad_map, mad_map + MAD_BUFSIZE - bytes_avail, bytes_avail);
    bytes_read = fread( mad_map + bytes_avail,sizeof(unsigned char), MAD_BUFSIZE - bytes_avail, mad_fd);
    map_offset += (MAD_BUFSIZE - bytes_avail);
    bytes_avail += bytes_read;
  }
}


int ReadMad::Decode( ) {
  
  struct mad_pcm *pcm;
  mad_fixed_t const *left_ch;
  mad_fixed_t const *right_ch;
  int nsamples;
  int nchannels;
  
  if ( bytes_avail < 3072) {
    //post("Filling buffer = %d,%d", bytes_avail,map_offset + MAD_BUFSIZE - bytes_avail);
    fill_buffer( -1);  // map_offset + MAD_BUFSIZE - bytes_avail);
    mad_stream_buffer(&stream, mad_map, bytes_avail);
  } else {
    //post("bytes_avail = %d", bytes_avail);
  }
  if (mad_frame_decode(&frame, &stream) == -1) {
    if (!MAD_RECOVERABLE( stream.error)) {
      mad_frame_mute(&frame);
      return 0;
    } else {
      if ( feof( mad_fd) ) {
	//post("MAD: end of file");
	return 0;
      }	
      //post("MAD error: (not fatal)"); 
    }
  }

  mad_synth_frame (&synth, &frame);
  
  {
    pcm = &synth.pcm;
    nsamples = pcm->length;
    nchannels = pcm->channels;
    left_ch = pcm->samples[0];
    right_ch = pcm->samples[1];
    int x = 0;
    while (nsamples--) {
      buffer[x++] = mad_f_tofloat(*(left_ch++));
      if (nchannels == 1) 
	buffer[x++] = mad_f_tofloat(*(left_ch-1));
      else /* nchannels == 2 */
	buffer[x++] = mad_f_tofloat(*(right_ch++));
      
    }
  }

  bytes_avail = stream.bufend - stream.next_frame;
  //if (pcm->length != 1152) 
  //post(" %d", pcm->length);
  return pcm->length*2;
}




ssize_t ReadMad::find_initial_frame(uint8_t *buf, int size) {
  uint8_t *data = buf;
  int ext_header = 0;
  int pos = 0;
  ssize_t header_size = 0;
  while (pos < (size - 10)) {
    if (pos == 0 && data[pos] == 0x0d && data[pos+1] == 0x0a)
      pos += 2;
    if (data[pos] == 0xff && (data[pos+1] == 0xfb
			      || data[pos+1] == 0xfa
			      || data[pos+1] == 0xf3 
			      || data[pos+1] == 0xe2
			      || data[pos+1] == 0xe3)) {
      //alsaplayer_error("found header at %d", pos);
      return pos;
    }	
    if (pos == 0 && data[pos] == 0x0d && data[pos+1] == 0x0a) {
      return -1; // Let MAD figure this out
    }	
    if (pos == 0 && (data[pos] == 'I' && data[pos+1] == 'D' && data[pos+2] == '3')) {
      header_size = (data[pos + 6] << 21) + 
	(data[pos + 7] << 14) +
	(data[pos + 8] << 7) +
	data[pos + 9]; /* syncsafe integer */
      if (data[pos + 5] & 0x10) {
	ext_header = 1;
	header_size += 10; /* 10 byte extended header */
      }

      header_size += 10;
      
      if (header_size > STREAM_BUFFER_SIZE) {
	//post("Header larger than 32K (%d)", header_size);
	return header_size;
      }	
      return header_size;
    } else if (data[pos] == 'R' && data[pos+1] == 'I' &&
	       data[pos+2] == 'F' && data[pos+3] == 'F') {
      pos+=4;
      //alsaplayer_error("Found a RIFF header");
      while (pos < size) {
	if (data[pos] == 'd' && data[pos+1] == 'a' &&
	    data[pos+2] == 't' && data[pos+3] == 'a') {
	  pos += 8; /* skip 'data' and ignore size */
	  return pos;
	} else
	  pos++;
      }
      puts("MAD debug: invalid header");
      return -1;
    } else if (pos == 0 && data[pos] == 'T' && data[pos+1] == 'A' && 
	       data[pos+2] == 'G') {
      return 128;	/* TAG is fixed 128 bytes, we assume! */
    }  else {
      pos++;
    }				
  }
  post("MAD debug: potential problem file or unhandled info block");
  post("next 4 bytes =  %x %x %x %x (index = %d, size = %d)",
	 data[header_size], data[header_size+1],
	 data[header_size+2], data[header_size+3],
	 header_size, size);
  return -1;
}


void ReadMad::Initialize( const char *filepath ) {
  char *p;

  if (( mad_fd = fopen(filepath, "r")) == NULL) {
    post( "Initialized() failed");
    return;  //return 0;
  }
  
  seekable = 1;  //assume seekable if its a file
  mad_synth_init  (&synth);
  mad_stream_init (&stream);
  mad_frame_init  (&frame);
  memset(&xing, 0, sizeof(struct xing));
  xing_init (&xing);
  mad_init = 1;
  fill_buffer(-1);
  //post("initial bytes_avail = %d", bytes_avail);
  offset = find_initial_frame( mad_map,
			       bytes_avail < STREAM_BUFFER_SIZE ? bytes_avail :
			       STREAM_BUFFER_SIZE);

  if ( offset < 0) {
    //post("mad_open() couldn't find valid MPEG header\n");
    offset = 0;
    //return false;
  }
  if ( offset > bytes_avail) {
    seekable = 1;
    //alsaplayer_error("Need to refill buffer (data->offset = %d)", data->offset);
    fill_buffer(0);
    mad_stream_buffer(&stream, mad_map, bytes_avail);
  } else {
    mad_stream_buffer(&stream, mad_map + offset, bytes_avail - offset);
    bytes_avail -= offset;
  }	
  
  if ((mad_frame_decode(&frame, &stream) != 0)) {
    //alsaplayer_error("MAD error: %s", error_str(data->stream.error, data->str));
    switch (stream.error) {
    case MAD_ERROR_BUFLEN:
      post( "MAD_ERROR_BUFLEN");
      return; //return 0;
    case MAD_ERROR_LOSTSYNC:
      if (mad_header_decode(&frame.header, &stream) == -1) {
	post("Invalid header (%s)", filepath);
      }
      mad_stream_buffer(&stream, stream.this_frame,
			bytes_avail - (stream.this_frame - mad_map));
      bytes_avail -= (stream.this_frame - mad_map);
      //post("avail = %d", data->bytes_avail);
      mad_frame_decode(&frame, &stream);
      break;
    case MAD_ERROR_BADBITALLOC:
      post( "MAD_ERROR_BADBITALLOC");
      return;//return 0;
    case MAD_ERROR_BADCRC:
      post("MAD_ERROR_BADCRC"); //error_str( stream.error, str));
    case 0x232:				
    case 0x235:
      break;
    default:
      post("No valid frame found at start (pos: %d, error: 0x%x --> %x %x %x %x) (%s)", 
	   offset, 
	   stream.error, 
	   stream.this_frame[0],
	   stream.this_frame[1],
	   stream.this_frame[2],
	   stream.this_frame[3],filepath);
      return;//return 0;
    }
  }
  if ( stream.error != MAD_ERROR_LOSTSYNC)
    if (xing_parse(&xing, stream.anc_ptr, stream.anc_bitlen) == 0) {
      // We use the xing data later on
    }	
  num_channels = ( frame.header.mode == MAD_MODE_SINGLE_CHANNEL) ? 1 : 2;
  samplerate = frame.header.samplerate;
  bitrate = frame.header.bitrate;
  mad_synth_frame (&synth, &frame);

  /* Calculate some values */
  bytes_avail = stream.bufend - stream.next_frame;
  {
    int64_t lframes;
    
    long oldpos = ftell( mad_fd);
    fseek( mad_fd, 0, SEEK_END);
    
    filesize = ftell(mad_fd);
    filesize -= offset;
    
    fseek( mad_fd, oldpos, SEEK_SET);
    if ( bitrate )
      time = ( filesize * 8) / (bitrate);
    else
      time = 0;
    
    samplesperframe = 32 * MAD_NSBSAMPLES(&frame.header);
    samplestotal = samplerate * (time+1);
    lframes = samplestotal / samplesperframe;
    nr_frames = xing.frames ? xing.frames : (int) lframes;
  }

  if ( nr_frames > 1000000 ) {
    seekable = 0; // Given really
  } else {
    seekable = 1;
  }	
  mad_init = 1;

  if (xing.frames) {
    post("xing.frames %d", xing.frames);
    seekable = 0;
  }


  return;//return 1;
}

bool ReadMad::Rewind() {
  if(mad_fd == NULL ) return false;
  seek_bytes( 0 );
  return true;
}

bool ReadMad::PCM_seek(long  position) {
  long byte_offset;
  if ( !seekable) return false;
  byte_offset = (long) ( ( (double) position / (double) samplestotal)  * filesize);
  seek_bytes( byte_offset );
  return true;
}



bool ReadMad::TIME_seek(double seconds) {
  //double time = ( filesize * 8) / (bitrate);
  long byte_offset;
  if ( !seekable) return false;
  byte_offset = (int) (( seconds / (double) time) * filesize);
  seek_bytes( byte_offset );
  return true;
}



void ReadMad::seek_bytes(long byte_offset) {
  struct mad_header header;
  int skip;
  mad_header_init(&header);
  
  bytes_avail = 0;
  skip = 0;
  //The total size in bytes for any given  frame can be calculated with the following formula: 
  //FrameSize = 144 * BitRate / (SampleRate + Padding).
  // 417.96 bytes: 144 * 128000 / (44100 + 0)
  if ( byte_offset > 4*418) {
    skip = 3;
  }
  //(seekpos / song length) * (filelength).

  //post("byte_offset %ld, position %ld, samplestotal %ld, filesize %ld", 
  //   byte_offset, position, samplestotal, filesize);
  fill_buffer( byte_offset );
  mad_stream_buffer(&stream, mad_map, bytes_avail);
  skip++;
  while (skip--) {
    mad_frame_decode(&frame, &stream);
    if (skip == 0)
      mad_synth_frame (&synth, &frame);
  }
  bytes_avail = stream.bufend - stream.next_frame;
  return;
}


///////////////// xing stuff


void ReadMad::xing_init(struct xing *xing)
{
  xing->flags = 0;
}

/*
 * NAME:	xing->parse()
 * DESCRIPTION:	parse a Xing VBR header
 */
int ReadMad::xing_parse(struct xing *xing, struct mad_bitptr ptr, unsigned int bitlen)
{
  if (bitlen < 64 || mad_bit_read(&ptr, 32) != XING_MAGIC)
    goto fail;

  xing->flags = mad_bit_read(&ptr, 32);
  bitlen -= 64;

  if (xing->flags & XING_FRAMES) {
    if (bitlen < 32)
      goto fail;

    xing->frames = mad_bit_read(&ptr, 32);
    bitlen -= 32;
  }

  if (xing->flags & XING_BYTES) {
    if (bitlen < 32)
      goto fail;

    xing->bytes = mad_bit_read(&ptr, 32);
    bitlen -= 32;
  }

  if (xing->flags & XING_TOC) {
    int i;

    if (bitlen < 800)
      goto fail;

    for (i = 0; i < 100; ++i)
      xing->toc[i] = mad_bit_read(&ptr, 8);

    bitlen -= 800;
  }

  if (xing->flags & XING_SCALE) {
    if (bitlen < 32)
      goto fail;

    xing->scale = mad_bit_read(&ptr, 32);
    bitlen -= 32;
  }

  return 0;

 fail:
  xing->flags = 0;
  return -1;
}


