#ifndef _READMAD_H_
#define _READMAD_H_

#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <inttypes.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>


#define MAD_BUFSIZE (32 * 1024)
#define XING_MAGIC    (('X' << 24) | ('i' << 16) | ('n' << 8) | 'g')

extern "C" { 	
#include <mad.h>
}

//#include "xing.h"
#include "Readsf.h"



#define STREAM_BUFFER_SIZE	(32 * 1024)
#define FRAME_RESERVE	2000


struct xing {
  long flags;			/* valid fields (see below) */
  unsigned long frames;		/* total number of frames */
  unsigned long bytes;		/* total number of bytes */
  unsigned char toc[100];	/* 100-point seek table */
  long scale;			/* ?? */
};

enum {
  XING_FRAMES = 0x00000001L,
  XING_BYTES  = 0x00000002L,
  XING_TOC    = 0x00000004L,
  XING_SCALE  = 0x00000008L
};

# define xing_finish(xing)	/* nothing */


class ReadMad : public Readsf {

 private:
  FILE *mad_fd;
  uint8_t mad_map[MAD_BUFSIZE];
  long map_offset;
  int bytes_avail;
  //stat stat;
  //ssize_t	*frames;
  //int highest_frame;
  //int current_frame;
  //char path[FILENAME_MAX+1];
  //char filename[FILENAME_MAX+1];
  struct mad_synth  synth; 
  struct mad_stream stream;
  struct mad_frame  frame;
  mad_timer_t timer;
  struct xing xing;

  int mad_init;
  ssize_t offset;
  ssize_t filesize;
  int samplerate;
  int bitrate;
  int samplesperframe;
  long samplestotal;
  long time;
  int seekable;
  int nr_frames;

  //float my_scale(mad_fixed_t sample);
  void fill_buffer(long newoffset);
  int mad_frame_seek( int frame );
  ssize_t find_initial_frame(uint8_t *buf, int size);
  void seek_bytes(long byte_offset);
  
  void xing_init(struct xing *);
  int xing_parse(struct xing *, struct mad_bitptr, unsigned int);

 public:
  ReadMad();
  ~ReadMad();
  virtual void Initialize(const char *file);
  virtual int Decode();
  virtual bool Rewind();
  virtual bool PCM_seek(long bytes);
  virtual bool TIME_seek(double seconds);
  
};		

#endif
